//
// This file is part of SerialDAS.
//
// SerialDAS is copyright (c) 2000 by
// Paul Gettings, Dep't of Geology & Geophysics, University of Utah
// All Rights Reserved.
// 
// SerialDAS is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 2, as
// published by the Free Software Foundation.
// 
// SerialDAS is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details, in the file
// COPYING.  If that file is not present, write to the Free
// Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
// 02111-1307, USA
// 
// The SerialDAS maintainers request that users forward any 
// improvements to <gettings@mines.utah.edu>.
//
package Utah.DAS.Serial;

import java.io.*;
import java.util.*;
//import javax.comm.*;
import gnu.io.*;

import Utah.DAS.Serial.SerialIO;
import Utah.DAS.Serial.PortParams;

/** Data acquisition from serial port instruments.
  * Based on the javax.comm classes from Sun; Linux version can use the RXTX
  * package for native methods.
  * <P>
  * Tested with Sun JDK 1.2.2 for linux, blackdown.org JDK 1.2.2 for
  * linux, IBM JDK 1.3 for linux, Sun JDK 1.3 for Win32
  * <P>
  * <UL>Limitations of the class:
  * <LI>Can only handle up to 4 ports in a single instance.
  * <LI>All instances set thread's priority to 2 below the highest possible.
  * </ul>
  * @author Paul Gettings, Dep't of Geology & Geophysics, University of Utah
  * @version 1.0.2 November 2000
  */
public class SerialDAS extends Thread {
  private SerialIO[] instrument;

  private boolean run;
  private Vector[] data;
  private boolean[] flushFlag;
  private boolean[] writeFlag;
  private String[] writeData;
  private int[] instrumentNums;

  private int runDelay = 25;

  private String[] shutdownString;

  private int nPorts;
  /** Maximum number of simultaneous open ports. Currently set to 4. */
  public final static int MAXINSTRUMENTS = 4;

  /** Constructor.
    * Initializes private variables for the instance.  After construction,
    * call openPort() for each instrument to be read.  Assuming all ports
    * open correctly, call start() to spawn a new thread.
    * <P>
    * This uses a thread to poll the instruments
    * and buffers records from them.  There is a
    * 25ms delay between attempted reads.
    * <P>
    * Call read(N) to get the most
    * recent record from instrument N.
    * <P>
    * Call end() to shutdown and close the ports.
    */
  public SerialDAS() {
    run = true;
    data = new Vector[MAXINSTRUMENTS];
    instrumentNums = new int[MAXINSTRUMENTS];
    instrument = new SerialIO[MAXINSTRUMENTS];
    shutdownString = new String[MAXINSTRUMENTS];
    writeData = new String[MAXINSTRUMENTS];
    writeFlag = new boolean[MAXINSTRUMENTS];
    flushFlag = new boolean[MAXINSTRUMENTS];
    nPorts = 0;
    for(int i=0; i<MAXINSTRUMENTS; i++) {
      writeFlag[i] = false;
      flushFlag[i] = false;
      data[i] = new Vector(10);
      }
    }

  /** Opens and sets up an instrument.
    * Opens the serial port to the instrument, sets the port up, and send the
    * startup string to the instrument.
    * <P>
    * Parity is encoded in a single integer:
    * <PRE>
    * 0    NONE
    * 1    EVEN
    * 2    ODD
    * 3    MARK
    * 4    SPACE
    * </pre>
    * <P>
    * The recordChars are characters that define the start and end of a single
    * data record.  Use '\0' to indicate any character.  This is not
    * a good idea for the end of record character, as that means the
    * record will be considered to end with any character; if set to
    * '\0', use a record length (see setRecordLength()).
    * @param N Number of the instrument; must be between [0, MAXINSTRUMENTS)
    * @param portName string name of desired port, as returned by getPortIdentifiers()
    * @param baud baud rate, in bps
    * @param dataBits # of data bits [5-8]
    * @param stopBits # of stop bits [1-2]
    * @param parity Encoded parity
    * @param recordChars[2] Array of delimiter characters for this port;
    * arranged as [start, end]
    * @param flowControl Encoded flow control setting for the port
    * @param startupString String to send to the instrument once port is open
    * @see Utah.DAS.Serial.SerialDAS#setRecordLength
    * @see Utah.DAS.Serial.SerialIO#setFlowControl
    */
  public boolean openPort(int N, String portName, int baud, int dataBits,
      int stopBits, int parity, char[] recordChars, int flowControl, String
      startupString, boolean debug) {
    if(N < 0 || N > MAXINSTRUMENTS-1)
      {
      if(debug) {
        System.err.println("openPort: Instrument number out of range.");
        }
      return false;
      }
    instrument[N] = new SerialIO(portName, baud, dataBits, stopBits, parity,
	recordChars);
    if(instrument[N].isOpen()) {
      if(instrument[N].setFlowControl(flowControl)) {
	if(instrument[N].write(startupString)) {
	  instrumentNums[nPorts] = N;
	  nPorts++;
	  shutdownString[N] = "";
	  return true;
	  }
	else if(debug) {
	  System.err.println("openPort: Error writing startup string.");
	  }
	}
      else if(debug) {
        System.err.println("openPort: Error setting flow control.");
        }
      }
    else if(debug) {
      System.err.println("openPort: Error opening port.");
      }
    return false;
    }

  /** Set the record length for instrument N.
    * @param N Number of instrument to set
    * @param length Length of record in bytes
    */
  public void setRecordLength(int N, int length) {
    if(N < 0 || N > MAXINSTRUMENTS) return;
    instrument[N].setRecordLength(length);
    }

  /** Set the record start/stop characters for instrument N.
    * @param N Number of instrument to set
    * @param recordChars Characters for start/stop of a record
    */
  public void setRecordChars(int N, char[] recordChars) {
    if(N < 0 || N > MAXINSTRUMENTS) return;
    instrument[N].setRecordChars(recordChars);
    }

  /** Set the shutdown string for an instrument.
    * @param N Number of instrument to set
    * @param msg String to send upon shutdown
    */
  public void setShutdownString(int N, String msg) {
    if(N < 0 || N > MAXINSTRUMENTS) return;
    shutdownString[N] = msg;
    }

  /** Set the sleep delay between polling loops.
    * Sets the sleep delay, in ms, between poll loops in the run()
    * method.  If N is negative, the delay is reset to the default 25
    * ms.
    * @param N Delay, in ms
    */
  public synchronized void setRunDelay(int N) {
    if(N < 0) {
      N = 25;
      }
    runDelay = N;
    }

  /** Continually reads the instruments and buffers new records.
    * <P>
    * On each loop iteration, the procedure is:
    * <OL>For each instrument-
    * <LI>Flush input, if it has been requested (flushInput())
    * <LI>Write a string, if it has been queued (write())
    * <LI>Read a new record, if available
    * </ol>
    * <P>
    * The loop has a delay of 25 ms
    * (default)
    * between loops to relinquish control for other threads to run.
    * Set the delay with setRunDelay().
    * <P>
    * Sets the read thread's priority to 2 below the highest.
    * <P>
    * Started by calling instance.start(); ended by calling instance.end()
    */
  public void run() {
    this.setPriority(Thread.MAX_PRIORITY-2);
    boolean flag;
    int i, N;
    String record;
    while(run) {
      for(i=0; i<nPorts; i++) {
	N = instrumentNums[i];

	// Flush input, if requested
	if(flushFlag[N]) {
	  instrument[N].flushInput();
	  flushFlag[N] = false;
	  }

	// write any pending output
	if(writeFlag[N]) {
	  instrument[N].write(writeData[N]);
	  writeFlag[N] = false;
	  }

	// read any pending input
	record = instrument[N].readPort();
	if(record != null) {
	  data[N].add(record);
	  }
	}
      try { Thread.sleep(runDelay); }
       catch (InterruptedException e) {}
      }
    for(i=0; i<nPorts; i++) {
      N = instrumentNums[i];
      instrument[N].write(shutdownString[N]);
      instrument[N].close();
      }
    }

  /** Stops the run() function, and hence the thread.  Call when shutting
    * down the system.  This will also close the serial ports cleanly.
    */
  public synchronized void end() {
    run = false;
    }

  /** Read the next record from instrument number N.
    * @return New record as a String, or null if no new record
    */
  public synchronized String read(int N) {
    if(N < 0 || N >= MAXINSTRUMENTS) {
      return null;
      }
    if(data[N].size() > 0) {
      return (String) data[N].remove(0);
      }
    else {
      return null;
      }
    }

    /** Write a String to an instrument.
      * Writes a String to the instrument, without any handshaking beyond the
      * hardware level.  Write is done in the next poll loop, before
      * the read.  Only the last String is buffered for write; wait for
      * a poll loop before writing another string.  Attempts to write
      * while output is pending will return false, and the pending output
      * is not overwritten.
      * @param N Number of instrument to write to
      * @param msg Message to be written
      * @return True on success
      */
    public synchronized boolean write(int N, String msg) {
      if(N < 0 || N >= MAXINSTRUMENTS) {
        return false;
        }
      if(writeFlag[N]) {
        return false;
        }
      writeFlag[N] = true;
      writeData[N] = msg;
      return true;
      }

    /** Flush input on an instrument.
      * This will discard any unread input.  It does not affect waiting
      * records that have already been buffered.  The flush is done in
      * the next poll loop, before a write or read.
      * @param N Number of instrument to flush
      */
    public synchronized void flushInput(int N) {
      if(N < 0 || N >= MAXINSTRUMENTS) {
        return;
        }
      flushFlag[N] = true;
      }

    /** Set input buffer size for a record.  The input buffer must be
      * large enough to hold an entire record at once, and should be
      * twice the size of a record.  The default buffer size is 64
      * bytes.
      * <p>
      * Calling this function destroys the old input buffer, losing any
      * data still in the buffer!  Therefore, care should be taken if
      * this function is called after DAS initialization.
      *
      * @param N Number of the instrument to reset the buffer size
      * @param nBytes Size of the new input buffer, in bytes
      */
    public synchronized void setInputBufferSize(int N, int nBytes) {
      if(N < 0 || N >= MAXINSTRUMENTS) {
        return;
        }
      instrument[N].setBufferSize(nBytes);
      }

} // end SerialDAS
