//
// This file is part of SerialDAS.
//
// SerialDAS is copyright (c) 2000 by
// Paul Gettings, Dep't of Geology & Geophysics, University of Utah
// All Rights Reserved.
// 
// SerialDAS is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 2, as
// published by the Free Software Foundation.
// 
// SerialDAS is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details, in the file
// COPYING.  If that file is not present, write to the Free
// Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
// 02111-1307, USA
// 
// The SerialDAS maintainers request that users forward any 
// improvements to <gettings@mines.utah.edu>.
//
package Utah.DAS.Serial;

import java.io.*;
import java.util.*;
//import javax.comm.*;
import gnu.io.*;

import Utah.DAS.Serial.PortParams;

/** Serial port interface class.  This class performs the actual read/write
  * operations on a port.  One instance is created for each port in use
  * by a SerialDAS instance.  Do not directly call this class; use the
  * methods in SerialDAS instead.
  * @author Paul Gettings, Dep't of Geology & Geophysics, University of Utah
  */
class SerialIO {
  private SerialPort portHandle;
  private static Enumeration portList;
  private static CommPortIdentifier portId;

  private int BUFFERSIZE = 64; // size of input record buffer, in bytes

  private int recordLength;

  private InputStream inputStream;
  private byte[] inputBuffer;
  private int bufOffset = 0;
  private OutputStream outputStream;

  private char endRecordChar;
  private char startRecordChar;

  private boolean portOpen = false;

  /** Constructor. Opens and sets up a serial port.
    * Parity is encoded in a single integer:
    * <PRE>
    * 0    NONE
    * 1    EVEN
    * 2    ODD
    * 3    MARK
    * 4    SPACE
    * </pre>
    * <P>
    * The recordChars are characters that define the start and end of a single
    * data record.  Use '\0' to indicate any character.  This is not
    * a good idea for the end of record character, as that means the
    * record will be considered to end with any character; if set,
    * use a minimum record length (see setRecordLength()).
    * @param portName string name of desired port, as returned by getPortIdentifiers()
    * @param baud baud rate, in bps
    * @param dataBits # of data bits [5-8]
    * @param stopBits # of stop bits [1-2]
    * @param parity Encoded parity
    * @param recordChars[2] Array of delimiter characters for this port;
    * arranged as [start, end]
    * @see Utah.DAS.Serial.SerialIO#setRecordLength
    */
  public SerialIO(String portName, int baud, int dataBits, int stopBits, int
      parity, char[] recordChars) {
    // setup variables
    inputBuffer = new byte[BUFFERSIZE];
    bufOffset = 0;

    // open the port
    portList = CommPortIdentifier.getPortIdentifiers();
    portOpen = false;
    while (portList.hasMoreElements()) {
      portId = (CommPortIdentifier) portList.nextElement();
      if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
	if (portId.getName().equals(portName)) {
	  portOpen = true;
	  try { portHandle = (SerialPort) portId.open("SerialDAS", 2000); }
	   catch (PortInUseException e) {portOpen=false;}
	  try { inputStream = portHandle.getInputStream(); }
	   catch (IOException e) {portOpen=false;}
	  try { outputStream = portHandle.getOutputStream(); }
	   catch (IOException e) {portOpen=false;}
	  // set the port parameters
	  if(!setParms(baud, dataBits, stopBits, parity)) portOpen=false;
	  // set input buffer size to large value, just in case
	  portHandle.setInputBufferSize(BUFFERSIZE);
	  // set a receive timeout - timeout of 0 indicates polling driver
	  // not event driven
	  try { portHandle.enableReceiveTimeout(50); }
	   catch (UnsupportedCommOperationException e) {portOpen=false;}
	  break;
	  }
	}
      }
    startRecordChar = recordChars[0];
    endRecordChar = recordChars[1];
  }

  /** Returns a list of the available port names.
    * @return String[] of port names
    */
  public static String[] getPortNames() {
    int n=0;
    portList = CommPortIdentifier.getPortIdentifiers();
    while (portList.hasMoreElements()) {
      portId = (CommPortIdentifier) portList.nextElement();
      if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
	n++;
	}
      }
    String[] portNames = new String[n];
    int i=0;
    portList = CommPortIdentifier.getPortIdentifiers();
    while (portList.hasMoreElements()) {
      portId = (CommPortIdentifier) portList.nextElement();
      if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
	portNames[i] = portId.getName();
	i++;
	}
      }
    return portNames;
    }

  /** Tests if a port is open and ready for use. */
  public boolean isOpen() {
    return portOpen;
    }

  /** Set the port parameters for the port
    * @param baud baud rate, in bps
    * @param dataBits # of data bits [5-8]
    * @param stopBits # of stop bits [1-2]
    * @param parity Coded as for openPort()
    * @return Returns true on success
    */
  public boolean setParms(int baud, int dataBits, int stopBits, int parity) {
    if(!portOpen) return false;
    switch(dataBits) {
      case 5: dataBits = SerialPort.DATABITS_5; break;
      case 6: dataBits = SerialPort.DATABITS_6; break;
      case 7: dataBits = SerialPort.DATABITS_7; break;
      case 8: dataBits = SerialPort.DATABITS_8; break;
      default: dataBits = SerialPort.DATABITS_8; break;
      }
    switch(stopBits) {
      case 1: stopBits = SerialPort.STOPBITS_1; break;
      case 2: stopBits = SerialPort.STOPBITS_2; break;
      default: stopBits = SerialPort.STOPBITS_1; break;
      }
    switch(parity) {
      case 0: parity = SerialPort.PARITY_NONE; break;
      case 1: parity = SerialPort.PARITY_EVEN; break;
      case 2: parity = SerialPort.PARITY_ODD; break;
      case 3: parity = SerialPort.PARITY_MARK; break;
      case 4: parity = SerialPort.PARITY_SPACE; break;
      default: parity = SerialPort.PARITY_NONE; break;
      }
    try { portHandle.setSerialPortParams(baud, dataBits, stopBits, parity); }
     catch (UnsupportedCommOperationException e) {return false;}
    return true;
    }

  /** Get port parameters.  Returns an instance of class PortParams which
    * has the baud rate, flow control, etc.
    */
  public PortParams getParms() {
    if(!portOpen) {return (new PortParams(0, 0, 0, 0));}
    int baud, stopBits, dataBits, parity;
    baud = portHandle.getBaudRate();
    parity = portHandle.getParity();
    stopBits = portHandle.getStopBits();
    dataBits = portHandle.getDataBits();
    switch(parity) {
      case SerialPort.PARITY_NONE: parity = 0; break;
      case SerialPort.PARITY_EVEN: parity = 1; break;
      case SerialPort.PARITY_ODD: parity = 2; break;
      case SerialPort.PARITY_MARK: parity = 3; break;
      case SerialPort.PARITY_SPACE: parity = 4; break;
      default: parity = -1; break;
      }
    switch(stopBits) {
      case SerialPort.STOPBITS_1: stopBits = 1; break;
      case SerialPort.STOPBITS_2: stopBits = 2; break;
      case SerialPort.STOPBITS_1_5: stopBits = 3; break;
      default: stopBits = 0; break;
      }
    switch(dataBits) {
      case SerialPort.DATABITS_5: dataBits = 5; break;
      case SerialPort.DATABITS_6: dataBits = 6; break;
      case SerialPort.DATABITS_7: dataBits = 7; break;
      case SerialPort.DATABITS_8: dataBits = 8; break;
      default: dataBits = 0; break;
      }
    return (new PortParams(baud, dataBits, stopBits, parity));
    }

  /** Set the flow control for the port.
    * No checking is done to make sure the port was successfully opened by
    * openPort().  Attempting to set flow control on an unopened port will
    * likely result in unhandled exception that terminates the JVM.
    * <P>
    * Input and output are set identically.
    * <P>
    * Flow control is encoded as a single integer:
    * <PRE>
    * 0 None
    * 1 XON/XOFF
    * 2 RTS/CTS
    * </pre>
    * @param flowcontrol Encoded flow control setting
    * @return Returns true on success
  */
  public boolean setFlowControl(int flowcontrol) {
    if(!portOpen) {return false;}
    switch(flowcontrol) {
      case 0: flowcontrol = SerialPort.FLOWCONTROL_NONE; break;
      case 1: flowcontrol = SerialPort.FLOWCONTROL_XONXOFF_IN |
        SerialPort.FLOWCONTROL_XONXOFF_OUT; break;
      case 2: flowcontrol = SerialPort.FLOWCONTROL_RTSCTS_IN |
	SerialPort.FLOWCONTROL_RTSCTS_OUT; break;
      default: flowcontrol = SerialPort.FLOWCONTROL_NONE; break;
      }
    try {portHandle.setFlowControlMode(flowcontrol); }
     catch (UnsupportedCommOperationException e) {return false;}
    return true;
    }

  /** Sets the size of the input buffer.  This should be larger than
    * any single record, and is recommended to be twice the length of a
    * record.
    * <p>
    * This function creates a new input buffer, which will lose the
    * contents of the old buffer!
    * @param nBytes Size of the input buffer, in bytes.
    */
  public void setBufferSize(int nBytes) {
    BUFFERSIZE = nBytes;
    inputBuffer = new byte[BUFFERSIZE];
    bufOffset = 0;
    }

  /** Sets the minimum length of a record to nBytes.  Set to 0 to indicate
    * any length record
    */
  public void setRecordLength(int nBytes) {
    if(nBytes > BUFFERSIZE) {
      this.setBufferSize(nBytes*2);
      }
    recordLength = nBytes;
    }

  /** Sets the record start/stop chars to recordChars.  Set to '\0' to indicate
    * no character.  If the start and stop character are not set, the
    * record length setting will determine when a "record" is ready for
    * read.
    * <p>
    * @param recordChars Start and stop characters that define a
    * record; '\0' indicates no character
    */
  public void setRecordChars(char[] recordChars) {
    startRecordChar = recordChars[0];
    endRecordChar = recordChars[1];
    }

  /** Reads the next record from the port.
    * @return String of the record; null indicates no record
    */
  public String readPort() {
    if(!portOpen) {return null;}
    int nBytes=0;
    try {
      int n;
      // read all available bytes in the buffer
      while(inputStream.available() > 0) {
        n = inputStream.read(inputBuffer, bufOffset,
	  inputBuffer.length-bufOffset);
	bufOffset += n;
	nBytes += n;
	if(bufOffset > BUFFERSIZE-1) {
	  bufOffset = BUFFERSIZE-1;
	  break;
	  }
	}
      }
     catch (IOException _) {return null;}
    // if no data (EOF), bail
    if(nBytes == -1) {
      return null;
      }
    // parse the inputBuffer for a record
    // parse as byte array, since faster than String ops!
    String record = null;
    int startIndex, endIndex;
    byte[] recordBytes;
    int i;
    startIndex = endIndex = -1;
    for(i=0; i<bufOffset; i++) {
      if(inputBuffer[i] == startRecordChar) {
	startIndex = i;
	}
      if(inputBuffer[i] == endRecordChar) {
	endIndex = i;
	break;
	}
      }
    if(startRecordChar == '\0') {
      startIndex = 0;
      }
    if(endRecordChar == '\0') {
      if(bufOffset >= recordLength) {
	endIndex = recordLength;
	}
      }
    // If we have a start character after an end character, then
    // no record!
    if(startIndex > endIndex) {
      startIndex = -1;
      }
    // If (startIndex & endIndex != -1), then we have a record.
    // Possible problem if we somehow miss the beginning of a record
    // in a previous read, clear the buffer, and then try for the next
    // record.
    if(startIndex != -1 && endIndex != -1) {
      recordBytes = new byte[endIndex-startIndex+1];
      for(i=startIndex; i<=endIndex; i++) {
	recordBytes[i-startIndex] = inputBuffer[i];
	}
      // convert to string
      record = new String(recordBytes);

      // Empty the input buffer of the record
      //System.out.println("*********");
      //System.out.println("readport:"+startIndex+" "+endIndex);
      //System.out.println("readport:"+bufOffset+"|"+(new String(inputBuffer)));
      for(i=endIndex+1; i<bufOffset; i++) {
	inputBuffer[startIndex+(i-endIndex-1)] = inputBuffer[i];
	}
      for(i=startIndex+(bufOffset-endIndex)-1; i<=bufOffset; i++) {
        if(i < 0) i = 0; // just in case
        inputBuffer[i] = 0;
        }
      bufOffset = startIndex + (bufOffset-endIndex) - 1;
      if(bufOffset < 0) { bufOffset = 0; } // just in case
      //System.out.println("readport:"+bufOffset+"|"+(new String(inputBuffer)));
      //System.out.println("*********");
      }

    return record;
    }

  /** Clears the port of pending input.
    * Reads available data on the port and tosses it.
    * Clears inputBuffer.
  */
  public void flushInput() {
    if(!portOpen) {return;}
    byte[] in;
    try {
      int nBytes = inputStream.available();
      inputStream.skip(nBytes);
      }
    catch (IOException _) {};
    // clear the inputBuffer
    inputBuffer = new byte[BUFFERSIZE];
    bufOffset = 0;
    }

  /** Write msg to the port.
    * @return True on success
    */
  public boolean write(String msg) {
    if(!portOpen) {return false;}
    try {outputStream.write(msg.getBytes());}
     catch (IOException _) {return false;}
     return true;
    }

  /** Close the port.
    * You should destroy this instance and create a new one to reopen
    * the port.
    */
  public void close() {
    portHandle.close();
    portOpen = false;
    }

} // end class SerialIO
